import numpy as np
import pandas as pd
import optimization as opt
import Activations as Act
import Losses as loss
from visualization import visualization as vis
from save_load import import_saved_weights,save_weights
import PIL, shutil, os


# encapsulate Layers parameters
class Layer:
    def __init__(self,size,activation='Identity'):
        '''
        Create a Layer with a specific size and activation function:

        Args:
            size: a tuble holds (input size, neuran count) respectively
            activation: specify the applied activation function on layer
                    avliable activations:
                        -Identity
                        -ReLU
                        -Sigmoid 
                        -Tanh 
                        -Softmax
                    Default: Identity

        Shape:
            - Input: size (N, m) N: number of features
            - Output: None

        Examples:
            layer1 = nn.Layer(size=(3,5), activation='ReLU')

        '''

        # check if the size forms of a tuple and activation as a string
        assert(type(size) == tuple)
        assert(type(activation) == str)

        self.inSize = size[0]
        self.outSize = size[1]
        # initialize weights and biases
        # weight of size (layer neurans no. , prev layer neurans no.)
        self.weights = np.random.randn(self.outSize,self.inSize) * np.sqrt( 2/(self.inSize+self.outSize) )
        # self.weights = np.random.randn(self.outSize,self.inSize)*0.01
        self.bias = np.zeros((1,self.outSize))

        # initialize weights and biases gradient
        self.weights_Grad = np.zeros_like(self.weights)
        self.bias_Grad = np.zeros_like(self.bias)

        # initialize A, Adash and Z where
        # Z:        linear output of layer
        # A:        layer output after activation 
        # Adash:   derivative of the layer output after activation
        self.Z = np.zeros_like(self.bias)
        self.A = np.zeros_like(self.bias)
        self.Adash = np.zeros_like(self.bias)

        self.activation = str.lower(activation)
    def forward(self,inputs):
        '''
        Calculate a forward propagation step:
        based on the Linear formula:
            Y = activation(X.W + b)

        Args:
            inputs: a numpy array with the previous layer values (or the network inputs)

        Return:
            a forward propagation step value after applying activation function

        Shape:
            - Input: inputs (1, N) N: number of features
            - Output: (1, m)

        Examples:
            layer = nn.Layer(size=(3,5), activation='ReLU')
            result = layer.forward(np.array([4,1,-5,3]))
        '''
        
        assert(type(inputs) == np.ndarray)
        assert(inputs.shape[1] == self.inSize)

        # linear w*x + b forward propagtion step
        self.Z = np.dot(inputs,self.weights.T) + self.bias
        # evaluate the results after the activation
        self.A, self.Adash = eval(f"Act.{self.activation}(Z = self.Z)")
        return self.A
        
    def __call__(self,inputs):
        '''
        Calculate a forward propagation step over one layer:
        based on the Linear formula:
            Y = activation(X.W + b)

        Args:
            inputs: a numpy array with the previous layer values (or the network inputs)

        Return:
            a forward propagation step value after applying activation function

        Shape:
            - Input: inputs (1, N) N: number of features
            - Output: (1, m)

        Examples:
            layer = nn.Layer(size=(3,5), activation='ReLU')
            result = layer(np.array([4,1,-5,3]))

        '''
        return self.forward(inputs)


# model class encapsulates Layers objects
class Model:
    def __init__(self,*layers):
        '''
        Create a Model with a specific number and type of Layers:

        Args:
            layers: multi-valued parameter that holds one or more Layer object

        Shape:
            - layers: (nn.Layer(N,m), nn.Layer(m,k), .... , nn.Layer(c,1))
            - Output: None

        Examples:
            model = nn.Model(
                Layer(size=(3,5), activation='ReLU'),
                Layer(size=(5,10), activation='ReLU'),
                Layer(size=(10,6), activation='ReLU'),
                Layer(size=(6,1), activation='ReLU')
                )

        '''

        assert(type(layers) == tuple)
        assert(type(layer) == "nn.Layer" for layer in layers)
        self.x = None
        self.y = None
        self.layers = list(layers)
    
    def forward(self,inputs):
        '''
        Calculate a forward propagation step over the whole model:
        based on the Linear formula:
            A = activation(Layer(x))
            B = activation(Layer(A))
            .
            .
            .
            Y = activation(Layer(F))

        Args:
            inputs: a numpy array with the network inputs

        Return:
            a forward propagation iteration value for the whole network
        
        Shape:
            - Input: inputs (1, N) N: number of features
            - Output: (1, m) m: number of nurans at output

        Examples:
            model = nn.Model(
                Layer(size=(3,5), activation='ReLU'),
                Layer(size=(5,1), activation='ReLU')
            )
            model.forward(np.array([1,6,-2]))
        '''
        
        assert(type(inputs) == np.ndarray)

        self.x = inputs
        for layer in self.layers:
            inputs = layer.forward(inputs)
        self.y = inputs
        return self.y

    def __call__(self,inputs):
        '''
        Calculate a forward propagation step over the whole model:
        based on the Linear formula:
            A = activation(Layer(x))
            B = activation(Layer(A))
            .
            .
            .
            Y = activation(Layer(F))

        Args:
            inputs: a numpy array with the network inputs

        Return:
            a forward propagation iteration value for the whole network
        
        Shape:
            - Input:  inputs (1, N) N: number of features
            - Output: (1, m) m: number of nurans at output

        Examples:
            model = nn.Model(
                Layer(size=(3,5), activation='ReLU'),
                Layer(size=(5,1), activation='ReLU')
            )
            model(np.array([1,6,-2]))
        '''

        return self.forward(inputs)


    def fit(self,dataset_input,label,optimization_type,loss_type,alpha,epoch,batch_size = 0 ,graph_on = False):
        '''
        Executing the learning process for a given dataset:
        Args:
            dataset_input: a numpy array with source inputs
            label: a numpy array with correct label for each example
            optimization_type: a string indicates the required learning optmization type
            loss_type: a string indicates the required output loss function to use
            alpha: a float number indicates the required learning rate
            epoch: an integer number required for number of iterations in learning process
            graph_on: a boolean typed value to visualize the process
        
        Return:
            None
        
        Shape:
            The shape has no changes, it just changes the current values for learning

        Examples:
            model.fit(X_train,label_train,'SGD','MSE',alpha = 0.0001,epoch = 50,graph_on = True)
        '''

        if(graph_on):
            self.graph = vis()
            if not os.path.exists('Loss'):
                os.mkdir("Loss")
        if len(label.shape) == 1:
            label = label.reshape(-1,1)
        if label.shape[1] == 1 and self.layers[-1].outSize != 1:
            tmp = np.zeros((dataset_input.shape[0],self.layers[-1].outSize))
            rows = np.arange(start = 0,stop = dataset_input.shape[0],step =1)
            label = np.squeeze(label)
            tmp[rows,label] = 1
            label = tmp
        counter = 0
        if(optimization_type == 'SGD'):
            while(True):
                counter += 1
                loss_acc = 0
                opt.init_delta(self)                                    
                for i in range(len(dataset_input)):
                    self.forward(dataset_input[i].reshape(1,-1))
                    loss_value , dloss =eval("loss."+str.lower(loss_type)+"(A = self.y , Y = label[i].reshape(1,-1))")
                    loss_acc += loss_value
                    opt.sgd(self,alpha,dataset_input[i].reshape(1,-1),dloss,len(dataset_input))
                if(graph_on):
                    self.graph.add_point_to_graph(loss_acc/len(dataset_input),counter,epoch)
                print(f"epoch: {counter},    Loss = {loss_acc/len(dataset_input)}")
                if(counter == epoch):
                    if(graph_on):
                        im = PIL.Image.open("Loss/0.png")
                        im.save('Loss.gif', save_all=True, append_images=[PIL.Image.open(f"Loss/{i}.png") for i in range(0,epoch)])
                        shutil.rmtree("Loss")
                    break
          #-------------------------------------------      
        elif(optimization_type == 'batch'):
              while(True):
                counter += 1
                loss_acc = 0
                opt.init_delta(self)
                for i in range(len(dataset_input)):
                    self.forward(dataset_input[i].reshape(1,-1))
                    loss_value , dloss = eval("loss."+str.lower(loss_type)+"(A = self.y , Y = label[i].reshape(1,-1))")
                    loss_acc += loss_value
                    opt.batch(self,dataset_input[i].reshape(1,-1),dloss)
                opt.update_weights_bias(self,alpha,len(dataset_input))
                if(graph_on):
                    self.graph.add_point_to_graph(loss_acc/len(dataset_input),counter,epoch)
                print(f"epoch: {counter},    Loss = {loss_acc/len(dataset_input)}")
                if(counter == epoch):
                    if(graph_on):
                        im = PIL.Image.open("Loss/0.png")
                        im.save('Loss.gif', save_all=True, append_images=[PIL.Image.open(f"Loss/{i}.png") for i in range(0,epoch)])
                        shutil.rmtree("Loss")
                    break
        #-------------------------------------------      
        elif(optimization_type == 'minibatch'):
            while(True):
                counter += 1
                loss_acc = 0
                opt.init_delta(self)
                for i in range(len(dataset_input)):
                    self.forward(dataset_input[i].reshape(1,-1))
                    loss_value , dloss = eval("loss."+str.lower(loss_type)+"(A = self.y , Y = label[i].reshape(1,-1))")
                    loss_acc += loss_value
                    opt.batch(self,dataset_input[i].reshape(1,-1),dloss)
                    if(i%batch_size == 0 or i == len(dataset_input)-1):
                        if(i != len(dataset_input)-1):
                            opt.init_delta(self)
                        opt.update_weights_bias(self,alpha,batch_size)                        
                if(graph_on):
                    self.graph.add_point_to_graph(loss_acc)
                print(f"epoch: {counter},    Loss = {loss_acc/len(dataset_input)}")
                if(counter == epoch):
                    if(graph_on):
                        im = PIL.Image.open("Loss/0.png")
                        im.save('Loss.gif', save_all=True, append_images=[PIL.Image.open(f"Loss/{i}.png") for i in range(0,epoch)])
                        shutil.rmtree("Loss")
                    break
                
            

    def evaluate(self,x,y,metric='Accuracy'):
        '''
        Calculate the evaluation matrices for the testing dataset
 
        Args:
            test_x: a numpy array with the network testing inputs
            test_y: a numpy array with the network testing true labels
            metric: metric name(s) as string or list of strings:
                    avliable metrics:
                        - Accuracy
                        - Confusion matrix
                        - Precision
                        - Recall
                        - F1 score
                Default: Accuracy
 
        Shape:
            - Input:
                test_x (K, N) K:number of testing samples, N: number of features
                test_y (K, m) K:number of testing samples, m: number of nurans at output
            - Output:
                the specified metric value(s)
                beside storing all metrics in model variables as follows:
                    self.accuracy [0-1] value
                    self.confusion_matrix (2,2) matrix
                    self.recall  [0-1] value
                    self.precision [0-1] value
                    self.f1_score [0-1] value
        Examples:
            P,R,F1 = model.evaluate(
                        np.array([[1,6,-2],[3,9,12],[7,-3,4]]),
                        np.array([[0],[1],[1]]),
                        metric=['Precision','Recall','F1_score']
                    )
        '''
        def HotForm(y,k):
            out = np.zeros((len(y),k))
            for i,e in enumerate(y):
                out[i][e] = 1
            return out
 
        assert(type(x) == np.ndarray)
        assert(type(y) == np.ndarray)
        assert(type(metric) == str or type(metric) == list)
 
        k = self.layers[-1].outSize
 
        # propagate in forward iteration over testing data
        y_tmp = self.forward(x)
        y_hat = np.array([np.argmax(t) for t in y_tmp])
 
        # convert a vector into a hot form
 
        TP = np.zeros((k))
        TN = np.zeros((k))
        FP = np.zeros((k))
        FN = np.zeros((k))
        y_hot = HotForm(y,k)
        y_hat_hot = HotForm(y_hat,k)
 
        for i in range(len(y)):
            for j in range(k):
                    TP[j] += (y_hot[i][j] == 1 and y_hat_hot[i][j] == 1)
                    FP[j] += (y_hot[i][j] == 0 and  y_hat_hot[i][j] == 1)
                    FN[j] += (y_hot[i][j] == 1 and  y_hat_hot[i][j] == 0)
                    TN[j] += (y_hot[i][j] == 0 and  y_hat_hot[i][j] == 0)
 
        percision = TP/(TP+FP+1e-9)
        recall = TP/(TP+FN+1e-9)
        f1_score = np.mean(2*(percision*recall)/(percision+recall+1e-9))
        percision = np.mean(percision)
        recall = np.mean(recall)
 
        confusion_matrix = np.zeros((k,k))
        for i,d in enumerate(y):
            confusion_matrix[y_hat[i],d] += 1
 
        accuracy = np.diag(confusion_matrix).sum()/confusion_matrix.sum()
 
        # return the targeted metric(s) 
        if type(metric) == str:
            metric = [metric]
        out = list()
        for m in metric:
            out.append(eval('_'.join(list(filter(lambda w: w != "" , str.lower(m).split(" "))))))
        return out

    def save(self,path = "model.NND"):
        '''
        Save model Object on Disk

        Args:
            Path to the model to be saved
                Default: "model.NND"

        Shape:
            - Input:
                path: string
            - Output:
                None
        Examples:
            model.save("my model.NND")
        '''
        save_weights(self,path)


def load(path):
    '''
    Import loaded model from Disk

    Args:
        Path to the model to be saved

    Shape:
        - Input:
            path: string
        - Output:
            model object loaded
    Examples:
        model = nn.load("my model.NND")
    '''
    return import_saved_weights(path)


import numpy as np
import pickle

def save_weights(model,path,csv_on= False):
    
    '''
        saves 2d array into csv file:
        Args:
            layer_arr: a 2d array holds numbers
        Shape:
            - Input: layer_arr (N, m) N: number of weights
            - Output: None
        Examples:
            save_weights(np.array([[1,2,3,4,56,9,7],[3,4,6,7,8]]))
        '''
    if csv_on:
        f = open('saved_weights.csv', 'w')
        for layer in model:
            f.write(",".join(str(x) for x in layer) + "\n")
    else:
        pickle.dump(model,open(path,'wb'))

def import_saved_weights(path,csv_on= False):
    
    '''
        import 2d array from csv file:
        Args: None
        Shape:
            - Input: None
            - Output: layer_arr (N, m) N: number of weights
        Examples:
            import_saved_weights()
    '''    
    if csv_on:
        all_of_the_weights = open(path, 'r').read()
        lines = all_of_the_weights.split('\n')
        layer_arr = []
        for line in lines:
            if len(line) > 1:
                x = line.split(',')
                xx = np.array(x)
                y = xx.astype(np.float)
                layer_arr.append(y)
        return layer_arr
    else:
        return pickle.load(open(path,'rb'))


# -*- coding: utf-8 -*-
"""
Created on Thu Jan 21 17:21:12 2021

@author: Mohamed Amr
"""

import pandas as pd
import math

class DataPreProcessing:
    
    def get_data(path, label_path = "", shuffle = False):
        '''
        Used to get feature matrix and label vector out of les.
        The feature matrix is a numpy array with dimensions (N X D), where N: number
        of samples and D: number of features in every sample.
        The label vector is a numpy array with dimensions (N X 1), where N: number
        of samples.

        • Parameters:
            1. path: {'string'} The path of the file that contains the features and
            labels.
            2. label path: {'string'} The path of labels file if label is not included
            in first path.
            3. shuffe: {'bool'} If True the rows will be rearranged randomly.
        • Return: {'tuple'} of 2 items
            1. Feature Matrix: {'numpy.array'}
            2. Label Vector: {'numpy.array'}
        • Limitations
            1. You must input label path if there are no labels data in first path.
            2. Label column name must be 'Label' and can't be anything else.
            '''
        df = pd.read_csv(path)
        df.columns = df.columns.map(str.lower)
        if(label_path != ""):
            df_temp = pd.read_csv(label_path)
            df_temp.columns = df_temp.columns.map(str.lower)
            df['label'] = df_temp['label']
        if(shuffle):
            df = df.sample(frac = 1)
        X = df.loc[:, df.columns != 'label'].to_numpy()
        label = df['label'].to_numpy()
        return (X, label)
        
    def normalize(matrix):
        '''
        Used to normalize input matrix, The returned matrix has the same dimen-
        sions as the input.
        • Parameters
            1. matrix: {'numpy.array'} Unormalized matrix.
        • Return: {'numpy.array'} normalized matrix.
        '''
        mean = matrix.mean()
        variance = matrix.var()
        return (matrix - mean)/math.sqrt(variance+0.0000001)
        
    def split_data(X, label):
        '''
        Used to split data into training data set with it's training labels and test
        data set with it's test labels.
        The data is split with splitting ratio 75% training data and 25% testing data.
        • Parameters
            1. X: {'numpy.array'} The feature matrix.
            2. label: {'numpy.array'} The labels vector.
        • Return: f'tuple'g of 4 items
            1. Training Feature Matrix: {'numpy.array'}
            2. Test Feature Matrix: {'numpy.array'}
            3. Training Label vector: {'numpy.array'}
            4. Test Label Vector: {'numpy.array'}
        '''
        ratio = int(len(label)*0.75)
        X_train = X[:ratio]
        X_test = X[ratio:]
        label_train = label[:ratio]
        label_test = label[ratio:]
        return (X_train, X_test, label_train, label_test)
    
################################################################################
        
#test cases:
#-----------
    
#data  = DataPreProcessing()
#X     = data.get_X('C:/Users/Mohamed Amr/Documents/Computer & Systems Engineering/4th CSE/First Term/Neural Networks/digit-recognizer/train.csv')
#label = data.get_label('C:/Users/Mohamed Amr/Documents/Computer & Systems Engineering/4th CSE/First Term/Neural Networks/digit-recognizer/train.csv')

# X, label = DataPreProcessing.get_data("C:/Users/Mohamed Amr/Documents/Computer & Systems Engineering/4th CSE/First Term/Neural Networks/digit-recognizer/train.csv", shuffle = True)
# X_train, X_test, label_train, label_test = DataPreProcessing.split_data(X, label)
# print("X_train: {}, Label_train: {}, X_test: {}, Label_test: {}".format(X_train.shape, X_test.shape, label_train.shape, label_test.shape))
# print(DataPreProcessing.normalize(X))



import matplotlib.pyplot as plt
from matplotlib import style
import os

class visualization:
    def __init__(self, name="the graph", line1="loss", line2="percision"):
        
        '''
        Create a graph with a specific name and line names if there is two lines:
        Args:
            name: a string holds the graph name
            line1: a string holds the first line name
            line2: a string holds the seconde line name
            
        Shape:
            - Input: graph and lines names
            - Output: None
        Examples:
            p1 = visualization("graph_title","first_line","seconde_line")
        '''

        self.value1 = []
        self.value2 = []
        self.value3 = []
        self.line1 = line1
        self.line2 = line2
        self.name = name

        style.use('fivethirtyeight')

        fig = plt.figure(name)
        self.ax1 = fig.add_subplot(1, 1, 1)
        plt.ion()
        plt.show()
        # plt.show(block = False)

    def add_point_to_graph(self,new_value3,it,epoch):
        
        '''
        update the graph by adding the new point to it and print it 
        and check if the final step reached we save the figure into picture and print it:
        Args:
            new_value3: a new value to be added to the graph
            it: an integer holds the number of the iteration
            epoch: an integer holds the number of the iteration witch we stop at
            
        Shape:
            - Input: three integer values
            - Output: None
        Examples:
            p1.add_point_to_graph(5,1,100)
        '''

        self.value3.append(new_value3)


        if len(self.value3) > 1 :
            self.ax1.clear()
            # self.ax1.plot(self.counter, self.value , c='black')
            self.ax1.plot( self.value3 )
            # plt.draw()
            plt.pause(10E-9)
            plt.savefig(f"Loss/{it-2}.png")
        if it == epoch:
            plt.savefig(f"final Loss.png")
            os.startfile("final Loss.png")
            
    def add_two_points_to_graph(self,new_value1,new_value2):
        
        '''
        update the graph by adding two new points to it then print it:
        Args:
            new_value1: a new value to be added to the graph
            new_value2: a new value to be added to the graph
            
        Shape:
            - Input: two integer values
            - Output: None
        Examples:
            p1.add_two_points_to_graph(10,4)
        '''
        
        self.value1.append(new_value1)
        self.value2.append(new_value2)


        if len(self.value1) > 1 :
            self.ax1.clear()
            self.ax1.plot( self.value1 ,label=self.line1)
            self.ax1.plot(self.value2,label=self.line2)
            plt.legend()
            # plt.draw()
            plt.pause(10E-9)



'''
date:23 january
Author: mohamed emad and mohamed amr mansi
'''
import numpy as np
def sgd(model,alpha,sample,dloss,size):
    '''
        makes one iteration on the given model using online optimization:
        based on the Linear formula:
            dl/dweights = d(loss)/d(activation_function n) * d(activation_function n)/d(prediction n) 
                * d(prediction n)/d(Input n) ........* d(activation_function 1)/d(prediction 1)* d(prediction 1)/d(Input_dataset)

        Args:
            -model : the structured model to be trained
            -alpha : the learning rate
            -sample: One sample from the dataset
            -dloss : drivative of the loss function by the activation function of the last layer 

        Return:
           -the function updates the weights and biases of the layers but doesn't return any thing
        
        Shape:
            - model:  object from the model class
            - alpha: scalar value
            - sample: np.array of shape(1 x number of features)
            - dloss:  np.array of shape( the same shape of the weights of the last layer)
        Examples:
            sgd(model , 0.1 , np.array([1 , 2 , 3 , 4]) , np.array([6])) '''
    for i in range(len(model.layers)):
         n = len(model.layers) - i -1
         if(i == 0): #output layer ==> there is no saved drivative value
             dloss_dact = dloss            
             dl_dw=np.dot(np.multiply(np.transpose(dloss_dact) ,np.transpose(model.layers[n].Adash)),model.layers[n-1].A)
             model.layers[n].weights_Grad += dl_dw
             dl_db=np.multiply(np.transpose(dloss_dact) ,np.transpose(model.layers[n].Adash))
             model.layers[n].bias_Grad += np.transpose(dl_db)
             saved_drevative = np.dot(model.layers[n].weights.transpose(),dl_db)
         else:       #hidden layers
             dactivation = np.multiply(saved_drevative ,np.transpose(model.layers[n].Adash))
             if(n == 0):
                 dl_dw=np.dot(dactivation,sample)
             else:
                 dl_dw=np.dot(dactivation,model.layers[n-1].A)
             model.layers[n].weights_Grad += dl_dw
             dl_db=dactivation
             model.layers[n].bias_Grad += np.transpose(dl_db)             
             saved_drevative = np.dot(model.layers[n].weights.transpose(),saved_drevative)        
         model.layers[n].weights=model.layers[n].weights-alpha*dl_dw/size
         model.layers[n].bias=model.layers[n].bias-alpha*np.transpose(dl_db)/size
         #print(n , model.layers[n].weights)
        
def batch(model,sample,dloss):
    '''
        makes one iteration on the given model using batch optimization:
        based on the Linear formula:
            dl/dweights = d(loss)/d(activation_function n) * d(activation_function n)/d(prediction n) 
                * d(prediction n)/d(Input n) ........* d(activation_function 1)/d(prediction 1)* d(prediction 1)/d(Input_dataset)

        Args:
            -model : the structured model to be trained
            -sample: One sample from the dataset
            -dloss : drivative of the loss function by the activation function of the last layer 

        Return:
           -the function accumilates the gredient of the loss wrt. weights and biases 
        
        Shape:
            - model:  object from the model class
            - sample: np.array of shape(1 x number of features)
            - dloss:  np.array of shape( the same shape of the weights of the last layer)
        Examples:
            batch(model , np.array([1 , 2 , 3 , 4]) , np.array([6])) '''    
    
    for i in range(len(model.layers)):
        n = len(model.layers) - i -1
        if(i == 0): #output layer ==> there is no saved drivative value
            dloss_dact = dloss            
            dl_dw=np.dot(np.multiply(np.transpose(dloss_dact) ,np.transpose(model.layers[n].Adash)),model.layers[n-1].A)
            model.layers[n].weights_Grad += dl_dw
            dl_db=np.multiply(np.transpose(dloss_dact),np.transpose(model.layers[n].Adash))
            model.layers[n].bias_Grad += np.transpose(dl_db)
            saved_drevative = np.dot(model.layers[n].weights.transpose(),dl_db)
        else:       #hidden layers
            dactivation = np.multiply(saved_drevative ,np.transpose(model.layers[n].Adash))
            if(n == 0):
                dl_dw=np.dot(dactivation,sample)
            else:
                dl_dw=np.dot(dactivation,model.layers[n-1].A)
            model.layers[n].weights_Grad += dl_dw
            dl_db=dactivation
            model.layers[n].bias_Grad += np.transpose(dl_db)             
            saved_drevative = np.dot(model.layers[n].weights.transpose(),saved_drevative)


def norm(model, size_of_dataset):
    '''
    this function used in the FIT function to get the  norm of the LASTLAYER to campere to eipslon  to
    stop the  while loop
    Args:
        model: to get from it the weights_Grad(delta)  of each layer
        size_of_dataset: length of dateset to make normaliztion to data

    Returns: void

    '''
    norms_weights = 0.0
    norms_bias = 0.0
    norms_weights = np.linalg.norm(model.layers[len(model.layers)-1].weights_Grad / size_of_dataset)
    norms_bias = np.linalg.norm(model.layers[len(model.layers)-1].bias_Grad / size_of_dataset)
# =============================================================================
#     for i in range(len(model.layers)):
#         norms_weights += np.linalg.norm(model.layers[i].weights_Grad / size_of_dataset)
#         norms_bias += np.linalg.norm(model.layers[i].bias_Grad / size_of_dataset)
# =============================================================================
    return norms_weights + norms_bias


def init_delta(model):
    '''
    in this function  i make zeros of all weights_Grad(delta) matrix in begin of
     optimations of every layer
    Args:
        model: to get from it the weights_Grad(delta) of each layer

    Returns:void

    '''
    for i in range(len(model.layers)):
        model.layers[i].weights_Grad = np.zeros_like(model.layers[i].weights_Grad)
        model.layers[i].bias_Grad = np.zeros_like(model.layers[i].bias_Grad)


def update_weights_bias(model, alpha, size_of_dataset):
    '''
     tis function is rule  to update  weights of each layer  according to that rule --> Wi+1=Wi-Alpha*gard
    Args:
        model:to get from it the weights_Grad(delta) and weights  of each layer
        alpha:Constant ->> learing Constant
        size_of_dataset:length of date set

    Returns:Void

    '''
    for i in range(len(model.layers)):
        model.layers[i].weights = model.layers[i].weights - alpha * (model.layers[i].weights_Grad / size_of_dataset)
        model.layers[i].bias = model.layers[i].bias - alpha * (model.layers[i].bias_Grad / size_of_dataset)



from setuptools import setup

with open("./NN-FRAMEWORK-FOR-DUMMIES/README.md","r") as fh:
    long_description = fh.read()
setup(
    name = "nn_for_dummies",
    version = '0.0.42',
    description = 'a small classfication neural network framework',
    py_modules  =['Activations','DataPreProcessing','Losses','nn','optimization','save_load','visualization'],
    package_dir = {'':'NN-FRAMEWORK-FOR-DUMMIES'},
    classifiers = ["Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Operating System :: OS Independent"
    ],
    long_description = long_description,
    long_description_content_type = "text/markdown",
)


import numpy as np

def mse(A,Y):
    '''
        Calculate the mean square loss and derivative of the mean sqaure loss between label and prediction :
        based on the formulas:
            Loss = (1/2m)*(sum((A-Y)^2))
            Loss_dash = (1/m)* (A-Y)
        Args:
            A: the activation output of the output layer
            Y: labels of the examples

        Return:
            Loss: The values of loss of the inputs
            Loss_dash:The values of the derivative of the loss related to the inputs
        
        Shape:
            - Input: A (m, k) m: number of samples , k: number of neurons in current layer,
                     Y (m, k) m: number of samples , k: number of neurons in current layer
            - Output: Loss,Loss_dash(m,k) m: number of samples , k: number of neurons in current layer

        Examples:
            result,result_der = mse(np.array([4,1,-5,3]),np.array([0,0,1,0]))
    '''
    m = A.shape[0]
    Loss = (1/(2*m))*np.sum(np.power(A-Y,2))
    Loss_dash = np.multiply((1/m),A-Y)
    return Loss,Loss_dash

def nll(A,Y):
    '''
        Calculate the negative log likelihood loss and derivative of negative log likelihood loss between label and prediction :
        based on the formulas:
            Loss = -log(prediction[yi]),yi is the index of the correct label
            Loss_dash = -1/Y
        Args:
            A: the activation output of the output layer
            Y: labels of the examples

        Return:
            Loss: The values of loss of the inputs
            Loss_dash:The values of the derivative of the loss related to the inputs
        
        Shape:
            - Input: A (m, k) m: number of samples , k: number of neurons in current layer,
                     Y (m, k) m: number of samples , k: number of neurons in current layer
            - Output: Loss,Loss_dash(m,k) m: number of samples , k: number of neurons in current layer

        Examples:
            result,result_der = nll(np.array([4,1,-5,3]),np.array([0,0,1,0]))
    '''
    Loss  = np.sum(-np.log(np.sum(np.multiply(A,Y),axis = 1,keepdims=True)))
    Loss_dash = np.divide(-1,Y)
    return Loss, Loss_dash

def l1(A,Y):
    '''
        Calculate the mean absolute loss and derivative of the mean absolute loss between label and prediction :
        based on the formulas:
            Loss = (1/m)*sum(|A-Y|)
            Loss_dash = 1 if A-Y > 0 otherwise -1
        Args:
            A: the activation output of the output layer
            Y: labels of the examples

        Return:
            Loss: The values of loss of the inputs
            Loss_dash:The values of the derivative of the loss related to the inputs
        
        Shape:
            - Input: A (m, k) m: number of samples , k: number of neurons in current layer,
                     Y (m, k) m: number of samples , k: number of neurons in current layer
            - Output: Loss,Loss_dash(m,k) m: number of samples , k: number of neurons in current layer

        Examples:
            result,result_der = l1(np.array([4,1,-5,3]),np.array([0,0,1,0]))
    '''
    m = A.shape[0]
    Loss = (1/(m))*np.sum(np.abs(A-Y))
    vf =  np.vectorize(lambda x: 1 if x > 0 else -1)
    Loss_dash = (1/(m))*vf(A-Y)
    return Loss,Loss_dash

def softmax(A,Y):
    '''
        Calculate the softmax loss and derivative of the softmax between label and prediction :
        based on the formulas:
            Loss = -log(prediction[yi]),yi is the index of the correct label
            Loss_dash[r] = -(1-yi) if r= yi otherwise yi
        Args:
            A: the activation output of the output layer
            Y: labels of the examples

        Return:
            Loss: The values of loss of the inputs
            Loss_dash:The values of the derivative of the loss related to the inputs
        
        Shape:
            - Input: A (m, k) m: number of samples , k: number of neurons in current layer,
                     Y (m, k) m: number of samples , k: number of neurons in current layer
            - Output: Loss,Loss_dash(m,k) m: number of samples , k: number of neurons in current layer

        Examples:
            result,result_der = softmax(np.array([4,1,-5,3]),np.array([0,0,1,0]))
    '''
    y = np.argmax(Y,axis = 1)
    Loss = -np.log([(A[i, y[i]]+1e-10) for i in range(len(A))])
    Loss[np.isnan(Loss)] = 1e-10
    Loss = np.mean(Loss)
    Loss_dash = (A - Y)/float(len(A))
    Loss_dash[np.isnan(Loss_dash)] = 1e-10
    return Loss,Loss_dash




import numpy as np

def sigmoid(Z):
    '''
        Calculate the sigmoid values and derivative of the sigmoid of the inputs :
        based on the formulas:
            A = 1/(1+e^(-Z))
            A_dash = A(1-A)
        Args:
            Z: a numpy array or matrix

        Return:
            A: The values of sigmoid of the input
            A_dash:The values of the derivative of the sigmoid related to the input
        
        Shape:
            - Input: Z (m, k) m: number of samples , k: number of neurons in current layer
            - Output: A(m, k),A_dash(m,k) m: number of samples , k: number of neurons in current layer

        Examples:
            result,result_der = sigmoid(np.array([4,1,-5,3]))
    '''
    A = np.divide(1,1+np.exp(-Z))
    A[np.isnan(A)] = 1e-10
    A_dash = np.multiply(A,1-A)
    A_dash[np.isnan(A_dash)] = 1e-10
    return A,A_dash

def tanh(Z):
    '''
        Calculate the tanh values and derivative of the tanh of the inputs :
        based on the formulas:
            A = (e^(Z)-e^(-Z))/(e^(Z)+e^(-Z))
            A_dash = 1-A^2
        Args:
            Z: a numpy array or matrix

        Return:
            A: The values of tanh of the input
            A_dash:The values of the derivative of the tanh related to the input
        
        Shape:
            - Input: Z (m, k) m: number of samples , k: number of neurons in current layer
            - Output: A(m, k),A_dash(m,k) m: number of samples , k: number of neurons in current layer

        Examples:
            result,result_der = tanh(np.array([4,1,-5,3]))
    '''
    A = np.tanh(Z)
    A[np.isnan(A)] = 1e-10
    A_dash = 1-np.power(A,2)
    return A,A_dash

def relu(Z):
    '''
        Calculate the relu values and derivative of the relu of the inputs :
        based on the formulas:
            A = Z if Z >= 0 otherwise 0 
            A_dash = 1 if Z >= 0 otherwise 0
        Args:
            Z: a numpy array or matrix

        Return:
            A: The values of relu of the input
            A_dash:The values of the derivative of the relu related to the input
        
        Shape:
            - Input: Z (m, k) m: number of samples , k: number of neurons in current layer
            - Output: A(m, k),A_dash(m,k) m: number of samples , k: number of neurons in current layer

        Examples:
            result,result_der = relu(np.array([4,1,-5,3]))
    '''
    A = Z*(Z >=0)
    A[np.isnan(A)] = 1e-10
    A_dash = 1.*(Z>=0)
    return A,A_dash

def softmax(Z):
    '''
        Calculate the softmax values and derivative of the softmax of the inputs :
        based on the formulas:
            A[i] = e^(Z[i])/sum(e^(Z[j])) j: from 1 to number of neruons in output layer
            A_dash = 1 , as it is included in the softmax loss function 
        Args:
            Z: a numpy array or matrix

        Return:
            A: The values of softmax of the input
            A_dash: 1
        
        Shape:
            - Input: Z (m, k) m: number of samples , k: number of neurons in current layer
            - Output: A(m, k),A_dash(m,k) m: number of samples , k: number of neurons in current layer

        Examples:
            result,result_der = softmax(np.array([4,1,-5,3]))
    '''
    A = np.exp(Z)
    A[np.isnan(A)] = 1e-10
    A = A/(np.sum(A,axis = 1,keepdims=True)+1e-10)
    A[np.isnan(A)] = 1e-10
    return A,1

def leakyrelu(Z):
    '''
        Calculate the leaky relu values and derivative of the leaky relu of the inputs :
        based on the formulas:
            A = Z if Z>= 0 otherwise Z*0.01
            A_dash = 1 if Z>= 0 otherwise 0.01
        Args:
            Z: a numpy array or matrix

        Return:
            A: The values of leaky relu of the input
            A_dash:The values of the derivative of the leaky relu related to the input
        
        Shape:
            - Input: Z (m, k) m: number of samples , k: number of neurons in current layer
            - Output: A(m, k),A_dash(m,k) m: number of samples , k: number of neurons in current layer

        Examples:
            result,result_der = leakyrelu(np.array([4,1,-5,3]))
    '''
    A = np.where(Z > 0, Z, Z * 0.01)
    A[np.isnan(A)] = 1e-10
    A_dash = Z
    A_dash[A_dash >= 0] = 1
    A_dash[A_dash < 0] = 0.01
    return A,A_dash

def identity(Z):
    '''
        Calculate the identity values and derivative of the identity of the inputs :
        based on the formulas:
            A = Z
            A_dash = 1
        Args:
            Z: a numpy array or matrix

        Return:
            A: The values of identity of the input
            A_dash:The values of the derivative of the identity related to the input
        
        Shape:
            - Input: Z (m, k) m: number of samples , k: number of neurons in current layer
            - Output: A(m, k),A_dash(m,k) m: number of samples , k: number of neurons in current layer

        Examples:
            result,result_der = identity(np.array([4,1,-5,3]))
    '''
    A = Z
    A[np.isnan(A)] = 1e-10
    A_dash = np.ones_like(A)
    return A,A_dash